"""
Internal utils
"""

from __future__ import annotations
from typing import Union, Optional, Type, TypeVar, Sequence, Tuple
import numpy

from returnn import frontend as _global_rf
from returnn.frontend._backend import Backend
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf

T = TypeVar("T")


def get_backend_from_tensors(*args):
    """
    :param args:
    :return: frontend, fallback to global frontend
    """
    for x in args:
        if isinstance(x, Tensor):
            # noinspection PyProtectedMember
            return x._raw_backend
    return _global_rf


def get_dtype_name(backend: Type[Backend], x: Union[T, Tensor[T], int, float]) -> str:
    """
    :param backend:
    :param x: tensor
    :return: dtype of tensor, as string
    """
    if isinstance(x, backend.RawTensorType):
        return backend.get_dtype_name_raw(x)
    elif isinstance(x, Tensor):
        return x.dtype
    elif isinstance(x, int):
        return rf.get_default_int_dtype()
    elif isinstance(x, float):
        return rf.get_default_float_dtype()
    else:
        raise TypeError(f"unexpected type {type(x)}")


def is_int(backend: Type[Backend], x: Union[T, Tensor[T], int, float]) -> bool:
    """
    :param backend:
    :param x:
    :return: whether the dtype is int
    """
    dtype = get_dtype_name(backend, x)
    return dtype.startswith("int") or dtype.startswith("uint")


def bin_op_out_template(
    backend: Type[Backend],
    a: Union[Tensor[T], int, float, numpy.number],
    b: Union[Tensor[T], int, float, numpy.number],
    *,
    name: str,
    res_dtype: Optional[str],
    allow_broadcast_all_sources: Optional[bool] = None,
    dim_order: Optional[Sequence[Dim]] = None,
) -> Tuple[Tensor[T], Tensor[T], Tensor[T]]:
    """
    make template for output tensor of binary op

    :param backend:
    :param a:
    :param b:
    :param name: str
    :param res_dtype: if not given, infer from a and b
    :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
        Not needed when out_dims is specified explicitly.
    :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
        Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
    :return: out, a, b
    """
    src_dtype = None
    if isinstance(a, Tensor):
        src_dtype = a.dtype
    elif isinstance(b, Tensor):
        src_dtype = b.dtype
    a = backend.convert_to_tensor(a, dtype=src_dtype)
    src_dtype = src_dtype or a.dtype
    b = backend.convert_to_tensor(b, dtype=src_dtype)
    # sanity checks
    # noinspection PyProtectedMember
    assert a._raw_backend == b._raw_backend, "Cannot combine tensors from two different frontends, e.g. TF and PT"
    assert a.dtype == b.dtype, f"For now only operations with Tensors of the same dtypes are supported, got {a} and {b}"
    all_dims = []
    for dim in a.dims + b.dims:
        if dim not in all_dims:
            all_dims.append(dim)
    if all(set(x.dims) != set(all_dims) for x in (a, b)):
        if allow_broadcast_all_sources is False:
            raise ValueError(f"compare: sources {a!r} {b!r} not allowed with allow_broadcast_all_sources=False")
        elif allow_broadcast_all_sources is None:
            raise ValueError(f"compare: sources {a!r} {b!r} require explicit allow_broadcast_all_sources=True")
        elif allow_broadcast_all_sources is True:
            pass
        else:
            raise TypeError(f"invalid type for allow_broadcast_all_sources: {type(allow_broadcast_all_sources)}")
    if dim_order:
        all_dims.sort(key=lambda d: dim_order.index(d) if d in dim_order else len(dim_order))
    if res_dtype is None:
        res_dtype = src_dtype
    out = Tensor(name, dims=all_dims, dtype=res_dtype)
    a = a.copy_compatible_to(out, check_dtype=False, check_sparse=False)
    b = b.copy_compatible_to(out, check_dtype=False, check_sparse=False)
    return out, a, b



"""
mark_as_output, mark_as_loss
"""

from __future__ import annotations
from typing import Optional, Sequence, TypeVar
from returnn.tensor import Tensor

T = TypeVar("T")

__all__ = ["mark_as_output", "mark_as_loss"]


def mark_as_loss(
    loss: Tensor,
    name: str,
    *,
    scale: Optional[float] = 1.0,
    as_error: bool = False,
    use_normalized_loss: bool = False,
    use_flatten_frames: bool = True,
    custom_inv_norm_factor: Optional[Tensor] = None,
) -> None:
    """
    Mark the given loss tensor as a loss.
    This has the effect that it is specially handled by RETURNN.
    Specifically, the optimizer can use it in training,
    and it is used for reporting per batch or per epoch,
    and for learning rate scheduling.

    This currently uses :class:`AsIsLoss` in RETURNN
    but this is an implementation detail and might change.

    :param loss:
    :param name: name of the loss. this name is used for reporting by RETURNN, and also for LR scheduling.
    :param scale: scale the loss by this factor for the training optimizer
      (but not for any reporting). setting to 0.0 has the effect that this loss is not used by the optimizer.
    :param as_error: if True, this loss is reported as an error instead of a loss,
      and not used by the training optimizer.
      This is by convention sth like the frame-error or edit-distance, and usually not differentiable anyway.
    :param bool use_flatten_frames: If True, will use :func:`returnn.tf.util.basic.flatten_with_seq_len_mask`,
      i.e. a "packed" sequence with the padded frames removed, and accumulates over that.
      This can be more efficient, also because it will further optimize incoming computations
      and e.g. skip softmax computations right before on the padded frames.
      This can also avoid issues with inf/nan in some cases.
      If False, it will mask the loss to 0 in the padded frames and accumulate over that.
      Typically, setting this to True (default) is both more efficient and better.
    :param bool use_normalized_loss: the loss used in optimization will be normalized.
      E.g. if the overall normalization is sum(loss)/sum(num_frames), this is also what the optimizer will use,
      otherwise the optimizer will just use sum(loss).
    :param custom_inv_norm_factor:
      The standard norm factor is 1/sum(target_seq_len) if the target has a time-axis,
      or 1/sum(output_seq_len) if there is no target and the output has a time-axis,
      or 1 otherwise. (See :func:`Loss.init` for details.)
      This is used for proper normalization of accumulated loss/error per epoch
      and also proper normalization per batch for reporting,
      no matter if use_normalized_loss is True or False.
      If you want to change this norm factor, you can set this.
      Basically, for all reporting, it uses sum(loss) * sum(custom_inv_norm_factor).
    """
    # noinspection PyProtectedMember
    loss._raw_backend.mark_as_loss(
        loss=loss,
        name=name,
        scale=scale,
        as_error=as_error,
        use_normalized_loss=use_normalized_loss,
        use_flatten_frames=use_flatten_frames,
        custom_inv_norm_factor=custom_inv_norm_factor,
    )


def mark_as_output(tensor: Tensor, name: str, *, shape: Optional[Sequence[int]] = None) -> None:
    """
    Mark this as an output.
    This has the effect that RETURNN will in any case construct the corresponding layer.
    Also see :func:`mark_as_default_output`.

    This is intended mostly for forwarding, or exporting the model (TF graph, TFLite, ONNX, etc).
    You must specify a shape to have the output shape (order of dims) well-defined
    (if not specified, we check if some defaults are possible, like BTF, or BF).

    :param tensor:
    :param name:
    :param shape: this specifies the order of the dims of the output, such that it is well-defined
        for some external application.
        If not specified, we try to infer BTF or BF as default, if that works, otherwise it will be an error.
    """
    # noinspection PyProtectedMember
    tensor._raw_backend.mark_as_output(tensor=tensor, name=name, shape=shape)



"""
High-level backend for RETURNN layers
"""

from __future__ import annotations
from typing import Union, Sequence, Optional, Any, Tuple, Dict
import numpy
import tensorflow as tf
import returnn.tf.compat as tf_compat

from returnn.util.basic import NotSpecified
from returnn.tensor import Tensor, Dim

# noinspection PyProtectedMember
from returnn.frontend._backend import Backend

from .layer import Layer
from .. import frontend_layers as rfl
from ... import frontend as rf
from . import dims as _dims
from returnn.frontend import RawTensorTypes


# Ignore this warning until we really expect that we implemented everything.
# noinspection PyAbstractClass
class ReturnnLayersBackend(Backend[Layer]):
    """
    RETURNN layers backend (using TF), where raw_tensor represents a RETURNN layer
    """

    RawTensorType = Layer
    is_tensorflow = True

    @staticmethod
    def executing_eagerly() -> bool:
        """executing eagerly"""
        return False

    @staticmethod
    def set_random_seed(seed: int):
        """
        :param seed:
        """
        tf_compat.v1.set_random_seed(seed)

    @staticmethod
    def get_random_state() -> Dict[str, bytes]:
        """
        :return: random state
        """
        # Not sure... we could cover all state vars used by RandomLayer...
        return {}

    @staticmethod
    def set_random_state(state: Dict[str, bytes]):
        """
        :param state: as returned by :func:`get_random_state`.
            This might not always be successful (e.g. different hardware, different backend version),
            so the calling code should always have called set_random_seed before to have the random generators
            in a reasonable fallback state.
        """
        assert not state  # not implemented... unexpected currently

    @staticmethod
    def get_dtype_name_raw(raw_tensor: Layer) -> str:
        """dtype"""
        return raw_tensor.tensor.dtype

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> str:
        """dtype"""
        return dtype_name

    @staticmethod
    def get_ndim_raw(raw_tensor: Layer) -> int:
        """ndim"""
        return raw_tensor.tensor.batch_ndim

    @staticmethod
    def get_shape_raw(raw_tensor: Layer) -> Layer:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: Layer) -> Tuple[Union[int, Layer]]:
        """shape"""
        raise NotImplementedError

    @staticmethod
    def get_known_shape_raw(raw_tensor: Layer) -> Tuple[Optional[int]]:
        """known shape"""
        return raw_tensor.tensor.batch_shape

    @staticmethod
    def set_known_shape_raw(raw_tensor: Layer, shape: Tuple[Optional[int]]) -> None:
        """set known shape"""
        pass  # ignore

    @staticmethod
    def fill_raw(shape: Union[Sequence[Union[int, Layer]], Layer], value: Union[Any, Layer]) -> Layer:
        """fill raw"""
        raise Exception("fill_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def compare_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """compare"""
        raise NotImplementedError  # TODO

    @staticmethod
    def combine_raw(a: Layer, kind: str, b: Layer) -> Layer:
        """combine"""
        raise NotImplementedError  # TODO

    @staticmethod
    def reshape_raw(raw_tensor: Layer, shape: Union[Sequence[Union[int, Layer]], Layer]) -> Layer:
        """reshape_raw"""
        raise Exception("reshape_raw not supported in layers backend because dim tags would be unknown")

    @staticmethod
    def transpose_raw(raw_tensor: Layer, perm: Sequence[int]) -> Layer:
        """transpose_raw is a no-op in this backend"""
        return raw_tensor

    @staticmethod
    def activation_raw(raw_tensor: Layer, func: str) -> Layer:
        """activation"""
        return rfl.make_layer(
            {"class": "activation", "activation": func, "from": raw_tensor.tensor}, name=func
        ).raw_tensor

    @staticmethod
    def softmax(tensor: Tensor, *, axis: Dim) -> Tensor:
        """softmax"""
        return rfl.make_layer({"class": "softmax_over_spatial", "axis": axis, "from": tensor}, name="softmax")

    @staticmethod
    def log_softmax(tensor: Tensor, *, axis: Dim) -> Tensor:
        """log softmax"""
        return rfl.make_layer(
            {"class": "softmax_over_spatial", "axis": axis, "from": tensor, "log_space": True}, name="log_softmax"
        )

    @staticmethod
    def sequence_mask_raw(lengths: Layer, *, batch_major: bool = True) -> Layer:
        """sequence mask"""
        raise NotImplementedError  # TODO

    @staticmethod
    def create_parameter_raw(tensor: rf.Parameter) -> Layer:
        """create parameter"""
        return rfl.make_layer(
            {"class": "variable", "shape": tensor.dims, "dtype": tensor.dtype}, name=tensor.name, existing_tensor=tensor
        ).raw_tensor

    @staticmethod
    def set_parameter_initial_value(param: rf.Parameter[Layer], value: Union[None, Tensor, rf.RawTensorTypes]) -> None:
        """set parameter initial value"""
        if value is None:
            param.raw_tensor.layer_dict.pop("init", None)
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
        elif isinstance(value, Tensor):
            param.raw_tensor.layer_dict.pop("init", None)
            if not value.raw_tensor.parent.can_access_children_from_root:
                accessible_parent = value.raw_tensor.parent
                while not accessible_parent.can_access_children_from_root:
                    accessible_parent = accessible_parent.parent
                value.raw_tensor.assign_parent(accessible_parent)
                # We could also maybe move out all the dependencies.
                # However, it's not clear whether this is always safe.
                for dep in value.raw_tensor.get_tensor_dependencies():
                    assert (
                        dep.parent.can_access_children_from_root
                    ), f"dep {dep} of moved value {value} is not accessible"
            param.raw_tensor.layer_dict["init_by_layer"] = value
        else:
            param.raw_tensor.layer_dict.pop("init_by_layer", None)
            param.raw_tensor.layer_dict["init"] = value
        if rfl.is_debug_eager_mode_enabled():
            shape = [d.get_dim_value() for d in param.dims]
            if isinstance(value, Tensor):
                assert value.placeholder is not None
                value_tf = value.placeholder
            else:
                value_tf = tf.broadcast_to(tf.convert_to_tensor(value), shape)
            if param.raw_tensor.debug_layer.output.placeholder is None:
                var = tf.Variable(value_tf, shape=[d.get_dim_value() for d in param.dims], dtype=param.dtype)
                param.raw_tensor.debug_layer.output.placeholder = var
            else:
                var = param.raw_tensor.debug_layer.output.placeholder
                assert isinstance(var, tf.Variable)
                var.assign(value_tf)

    @staticmethod
    def convert_to_tensor(
        value: Union[Tensor, Layer, RawTensorTypes],
        *,
        dims: Sequence[Dim] = (),
        dtype: Optional[str] = None,
        sparse_dim: Optional[Dim] = None,
    ) -> Tensor[Layer]:
        """convert to tensor"""
        if isinstance(value, Tensor):
            assert value.dims == tuple(dims)
            if dtype:
                assert value.dtype == dtype
            return value
        kwargs = {}
        dim_deps = _dims.get_dim_deps(dims)
        if dim_deps:
            kwargs["shape_deps"] = dim_deps
        return rfl.make_layer(
            {"class": "constant", "value": value, "shape": dims, "dtype": dtype, **kwargs}, name="constant"
        )

    @staticmethod
    def matmul(a: Tensor, b: Tensor, *, reduce: Union[Dim, Sequence[Dim]]) -> Tensor:
        """matmul"""
        return rfl.make_layer({"class": "dot", "from": [a, b], "reduce": reduce}, name="matmul")

    @staticmethod
    def range_over_dim(dim: Dim) -> Tensor:
        """range over dim"""
        return rfl.make_layer(
            {"class": "range_in_axis", "from": _dims.get_dim_deps(dim), "axis": dim}, name="range_over_dim"
        )

    @staticmethod
    def reduce(
        source: Tensor, *, mode: str, axis: Union[Dim, Sequence[Dim]], use_time_mask: bool = NotSpecified
    ) -> Tensor:
        """Reduce"""
        assert mode in Backend._AllowedReduceModes
        kwargs = {}
        if use_time_mask is not NotSpecified:
            kwargs["use_time_mask"] = use_time_mask
        return rfl.make_layer(
            {"class": "reduce", "from": source, "mode": mode, "axis": axis, **kwargs}, name=f"reduce_{mode}"
        )

    @staticmethod
    def random(
        *,
        dims: Sequence[Dim],
        dtype: str,
        sparse_dim: Optional[Dim] = None,
        distribution: str,
        mean: Optional[Union[int, float, Tensor]] = None,
        stddev: Optional[Union[int, float, Tensor]] = None,
        bound: Optional[Union[int, float, Tensor]] = None,
        minval: Optional[Union[int, float, Tensor]] = None,
        maxval: Optional[Union[int, float, Tensor]] = None,
        seed: Optional[Union[int, Sequence[int], numpy.ndarray]] = None,
        algorithm: Optional[str] = None,
        explicit_state: Optional[Tensor] = None,
        auto_update_state: Optional[bool] = None,
        static: Optional[bool] = None,
    ) -> Tensor:
        """random"""
        kwargs = {
            "mean": mean,
            "stddev": stddev,
            "bound": bound,
            "minval": minval,
            "maxval": maxval,
            "seed": seed,
            "algorithm": algorithm,
            "explicit_state": explicit_state,
            "auto_update_state": auto_update_state,
            "static": static,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return rfl.make_layer(
            {
                "class": "random",
                "shape": dims,
                "dtype": dtype,
                "sparse_dim": sparse_dim,
                "distribution": distribution,
                **kwargs,
            },
            name="random",
        )



"""
RETURNN frontend (returnn.frontend) utils
"""

from __future__ import annotations
from typing import Callable
import contextlib

from returnn.config import Config, global_config_ctx
import returnn.frontend as rf
from returnn.tensor import Tensor, TensorDict
import returnn.tf.compat as tf_compat
from returnn.tf.network import TFNetwork, ExternData


@contextlib.contextmanager
def tf_scope():
    """tf scope"""
    with tf_compat.v1.Graph().as_default(), tf_compat.v1.Session().as_default() as session:
        yield session


def run_model(get_model: Callable[[], rf.Module], extern_data: Tensor):
    """run"""
    out_pt = run_model_torch(get_model, extern_data)
    out_tf = run_model_net_dict_tf(get_model, extern_data)
    out_pt, out_tf  # noqa  # TODO ...


def run_model_torch(get_model: Callable[[], rf.Module], extern_data: Tensor) -> Tensor:
    """run"""
    rf.select_backend_torch()
    rf.set_random_seed(42)
    _fill_random(extern_data)
    model = get_model()
    out = model(extern_data)
    assert isinstance(out, Tensor)
    return out


def run_model_net_dict_tf(get_model: Callable[[], rf.Module], extern_data: Tensor):
    """run"""
    rf.select_backend_returnn_layers_tf()
    rf.set_random_seed(42)
    extern_data_name = extern_data.name

    # noinspection PyUnusedLocal
    def _get_model(*, epoch: int, step: int) -> rf.Module:
        return get_model()

    # noinspection PyShadowingNames
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = model(extern_data.data[extern_data_name])
        out.mark_as_default_output(shape=out.dims)

    extern_data_ = ExternData()
    extern_data_.register_data(extern_data)
    config = Config()

    from returnn.tf.frontend_layers.config_entry_points import get_net_dict

    with tf_scope() as session, global_config_ctx(config):
        _fill_random(extern_data)

        net_dict = get_net_dict(
            epoch=1, step=0, get_model_func=_get_model, extern_data=extern_data_, step_func=_forward_step
        )

        net = TFNetwork(config=config, extern_data=extern_data_, train_flag=False)
        net.construct_from_dict(net_dict)
        out = net.get_default_output_layer().output

        session.run(tf_compat.v1.global_variables_initializer())
        out_v = session.run(out.placeholder)
        out = out.copy_template()
        out.raw_tensor = out_v
        return out


def _fill_random(x: Tensor) -> bool:
    """fill. return whether sth was filled"""
    raw_tensor_type = rf.get_raw_tensor_type()
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if not dim.dyn_size_ext:
                continue
            if _fill_random(dim.dyn_size_ext):
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True
            elif not isinstance(dim.dyn_size_ext.raw_tensor, raw_tensor_type):
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, raw_tensor_type):
            x.raw_tensor = None

    if x.raw_tensor is None:
        if x.dtype.startswith("int"):
            min_val = 0
            max_val = 10
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            x.raw_tensor = rf.random(
                dims=x.dims,
                dtype=x.dtype,
                sparse_dim=x.sparse_dim,
                minval=min_val,
                maxval=max_val,
                distribution="uniform",
            ).raw_tensor
        elif x.dtype.startswith("float"):
            x.raw_tensor = rf.random(
                dims=x.dims, dtype=x.dtype, sparse_dim=x.sparse_dim, distribution="normal", mean=0.0, stddev=1.0
            ).raw_tensor
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        filled = True

    assert isinstance(x.raw_tensor, raw_tensor_type)

    return filled



"""
Dot / matmul
"""


from __future__ import annotations
from typing import Sequence, Union, TypeVar
from returnn.tensor import Tensor, Dim

T = TypeVar("T")

__all__ = ["matmul", "dot"]


# noinspection PyShadowingNames
def matmul(a: Tensor[T], b: Tensor[T], *, reduce: Union[Dim, Sequence[Dim]]) -> Tensor[T]:
    """
    This performs a batched matmul of two sources a and b
    (non-batched matmul and dot product are special cases).
    The underlying operation is a batched matmul (shared..., I, J) * (shared..., J, K) -> (shared..., I, K).
    The inputs a and b are transformed internally into the required shapes in the following way:
    The axis J is specified via the Dim given as 'reduce'. If multiple reduce Dims are given the corresponding axes
    are merged into one before the matmul via a reshape. All other matching Dims in a and b will be treated as
    batch dimensions ('shared...'). Dims unique to a and b define the axes I and K, respectively. (Multiple or no
    unique axes in a and b are supported too.)

    Depending on which Dims exist in a, b and reduce this dot operation can be used to compute scaling, scalar
    product, outer product, matrix-vector multiplication, matrix-matrix multiplication etc. (all possibly batched).

    :param a:
    :param b:
    :param reduce: Dims over which to perform the product, have to be present in both a and b
    :return: result of dot product, Dim order: common axes as sorted in a, unique axes of a (in order),
        unique axes of b (in order)
    """
    # noinspection PyProtectedMember
    return a._raw_backend.matmul(a=a, b=b, reduce=reduce)


# alias for some older code
dot = matmul



"""
Types
"""

from __future__ import annotations
from typing import Union, Type, TYPE_CHECKING
import numpy
from returnn.tensor import Tensor, Dim
from ._backend import global_backend


__all__ = ["RawTensorTypes", "Tensor", "Dim", "get_raw_tensor_type"]


RawTensorTypes = Union[int, float, complex, numpy.number, numpy.ndarray, bool, str]


def get_raw_tensor_type() -> Type:
    """
    :return: the raw tensor type of the current selected backend, e.g. ``torch.Tensor`` or ``tf.Tensor``
    """
    if TYPE_CHECKING:
        import torch

        return torch.Tensor  # just as an example
    return global_backend.RawTensorType
